import pytest

import numpy as np
from cosapp.ports import Port
from cosapp.systems import System

from cosapp.tests.library.systems.multiply import IterativeNonLinear


class Float(Port):
    def setup(self):
        self.add_variable('x', 1.)


class AllCausality(System):
    def setup(self):
        self.add_input(Float, "in_")
        self.add_inward("in_x", 22.)
        self.add_output(Float, "out")
        self.add_outward("out_y", 42.)

@pytest.fixture(scope='function')
def allcausality():
    return AllCausality('test')


class TestType(System):

    def setup(self):
        self.add_inward("a", True)
        self.add_inward("b", 22)
        self.add_inward("c", 2.54)
        self.add_inward("d", "string")
        self.add_inward("f", np.array(True))
        self.add_inward("g", np.array(22))
        self.add_inward("h", np.array(2.54))
        self.add_inward("i", np.array("string"))

        self.add_inward("k", np.array([0., 1, 2, ]))
        self.add_inward("l", dict(a=1, b=2))

@pytest.fixture(scope='function')
def testtype():
    return TestType('test')


class ExpRampOde(System):
    """
    System representing function f(t) = a * (1 - exp(-t / tau)),
    through ODE: tau * f' + f = a
    """
    def setup(self):
        self.add_inward('a', 1.0)
        self.add_inward('tau', 1.0)

        self.add_outward('df_dt', 0.0)
        self.add_transient('f', der='df_dt', max_time_step='tau / 5')

    def compute(self):
        self.df_dt = (self.a - self.f) / self.tau

    def __call__(self, t):
        """Analytical solution at time t"""
        return self.a * (1 - np.exp(-t / self.tau))

@pytest.fixture(scope='function')
def ode():
    return ExpRampOde('ode')


class VectorSyst(System):

    def setup(self):
        self.add_inward("vi", [1, 2, 3])
        self.add_inward("vr", [1., 2., 3.])

@pytest.fixture(scope='function')
def vector_syst():
    return VectorSyst("syst")


@pytest.fixture(scope='function')
def iterativenonlinear():
    snl = IterativeNonLinear('nl')
    snl.splitter.inwards.split_ratio = 0.1
    snl.mult2.inwards.K1 = 1
    snl.mult2.inwards.K2 = 1
    snl.nonlinear.inwards.k1 = 1
    snl.nonlinear.inwards.k2 = 0.5

    return snl


class VectorProblem(System):

    def setup(self):
        self.add_inward("in_")
        self.add_outward("out")

        self.add_inward("dummy_coef", np.array([1., 2.]))

        self.add_unknown("dummy_coef[0]").add_equation("dummy_coef[0] == 0.5 * (in_ + out)")

    def compute(self):
        self.out = self.dummy_coef[1] + self.in_

@pytest.fixture(scope='function')
def vector_problem():
    return VectorProblem("problem")



import pytest
import logging
import re
import weakref
from unittest import mock

from cosapp.core.module import Module
from cosapp.core.signal import Slot
from cosapp.drivers import Driver
from cosapp.utils.logging import LogFormat, LogLevel
from cosapp.utils.orderedset import OrderedSet
from cosapp.utils.testing import no_exception


@pytest.fixture(autouse=True)
def PatchModule():
    """Patch Module to make it instanciable for tests"""
    patcher = mock.patch.multiple(
        Module,
        __abstractmethods__ = set(),
        is_standalone = lambda self: False,
    )
    patcher.start()
    yield
    patcher.stop()


@pytest.fixture(autouse=True)
def Russian_dolls():
    def factory(name: str, *submodules: str):
        head = Module(name)
        module_list = [head]
        last = head
        for name in submodules:
            last = last.add_child(Module(name))
            module_list.append(last)
        return module_list

    return factory


@pytest.fixture(scope="function")
def fake():
    return Module("fake")


@pytest.mark.parametrize("name, error", [
    ("a", None),
    ("A", None),
    ("foobar", None),
    ("foo4bar", None),
    ("loveYou2", None),
    ("CamelBack", None),
    ("foo_bar", None),
    ("foobar_", None),
    ("_foobar", ValueError),
    ("foo bar", ValueError),
    ("foobar?", ValueError),
    ("foo.bar", ValueError),
    ("foo:bar", ValueError),
    ("foo/bar", ValueError),
    ("1foobar", ValueError),
    ("foobar-2", ValueError),
    ("foobar:2", ValueError),
    ("foobar.2", ValueError),
    ("foo.bar", ValueError),
    ("inwards", ValueError),
    ("outwards", ValueError),
    (23, TypeError),
    (1.0, TypeError),
    (dict(a=True), TypeError),
    (list(), TypeError),
])
def test_Module__init__(name, error):
    if error is None:
        module = Module(name)
        assert module.name == name
    else:
        with pytest.raises(error):
            Module(name)


def test_Module__weakref__(fake):
    with no_exception():
        proxy = weakref.proxy(fake)


def test_Module_name_setter():
    module = Module("foo")
    assert module.name == "foo"
    module.name = "bar"
    assert module.name == "bar"
    with pytest.raises(ValueError):
        module.name = "invalid.name"


def test_Module_contextual_name(Russian_dolls):
    a, b, c, d = Russian_dolls(*iter('abcd'))

    assert a.contextual_name == ""
    assert b.contextual_name == "b"
    assert c.contextual_name == "b.c"
    assert d.contextual_name == "b.c.d"


def test_Module_fullname(Russian_dolls):
    a, b, c, d = Russian_dolls(*iter('abcd'))

    assert a.full_name(trim_root=True) == ""
    assert b.full_name(trim_root=True) == "b"
    assert c.full_name(trim_root=True) == "b.c"
    assert d.full_name(trim_root=True) == "b.c.d"

    assert a.full_name() == "a"
    assert b.full_name() == "a.b"
    assert c.full_name() == "a.b.c"
    assert d.full_name() == "a.b.c.d"


def test_Module_path_namelist(Russian_dolls):
    a, b, c, d = Russian_dolls(*iter('abcd'))

    assert a.path_namelist() == ["a"]
    assert b.path_namelist() == ["a", "b"]
    assert c.path_namelist() == ["a", "b", "c"]
    assert d.path_namelist() == ["a", "b", "c", "d"]


def test_Module_compute_calls():
    m = Module("m")

    assert m.compute_calls == 0

    m.run_once()
    assert m.compute_calls == 1

    m.call_setup_run()  # Reset counter
    assert m.compute_calls == 0


def test_Module_size():
    s = Module("s")
    T = Module("T")
    u = Module("u")
    v = Module("v")

    s.add_child(T)
    T.add_child(u)
    T.add_child(v)
    s.add_child(Module("r"))

    assert s.size == 5
    assert u.size == 1
    assert T.size == 3


def test_Module_get_path_to_child(Russian_dolls):
    a, b, c, d = Russian_dolls("a", "b", "c", "d")

    assert a.get_path_to_child(a) == ""
    assert a.get_path_to_child(b) == "b"
    assert a.get_path_to_child(c) == "b.c"
    assert a.get_path_to_child(d) == "b.c.d"
    assert b.get_path_to_child(c) == "c"
    assert b.get_path_to_child(d) == "c.d"

    with pytest.raises(ValueError, match="not a child of 'b'"):
        b.get_path_to_child(a)

    orphan = Module('orphan')
    with pytest.raises(ValueError, match="not a child of 'a'"):
        a.get_path_to_child(orphan)


def test_Module_pop_child_driver():
    # From Driver
    d = Driver("dummy")
    assert d.children == {}

    update = d.add_child(Driver("mydriver"))
    assert d.children == {"mydriver": update}

    d.pop_child("mydriver")
    assert d.children == {}

    with pytest.raises(AttributeError):
        d.pop_child("mydriver")


@pytest.mark.parametrize("values, expected", [
    ([], dict(values=[])),
    (None, dict(values=[])),
    (['foo', 'bar'], dict(values=['foo', 'bar'])),
    (['foo', 'bar', 'foo'], dict(values=['foo', 'bar'])),
    (('foo', 'bar', 'foo'), dict(values=['foo', 'bar'])),
    ([1, 'foo'], dict(error=TypeError, match="All elements of .*exec_order must be strings")),
    (2, dict(error=TypeError, match="object is not iterable")),
])
def test_Module_exec_order(fake, values, expected):
    error = expected.get("error", None)
    if error is None:
        fake.exec_order = values
        assert isinstance(fake.exec_order, OrderedSet)
        assert list(fake.exec_order) == expected["values"]
    else:
        pattern = expected.get("match", None)
        with pytest.raises(error, match=pattern):
            fake.exec_order = values


def test_Module_setup_ran(fake):
    with mock.patch("cosapp.core.signal.signal.inspect"):
        fake_callback = mock.Mock(spec=lambda **kwargs: None)
        fake_callback.return_value = None

        fake.setup_ran.connect(Slot(fake_callback))
        fake.call_setup_run()

        fake_callback.assert_called_once_with()


def test_Module_computed(fake):
    with mock.patch("cosapp.core.signal.signal.inspect"):
        fake_callback = mock.Mock(spec=lambda **kwargs: None)
        fake_callback.return_value = None

        fake.computed.connect(Slot(fake_callback))
        fake.run_once()

        fake_callback.assert_called_once_with()


def test_Module_clean_ran(fake):
    with mock.patch("cosapp.core.signal.signal.inspect"):
        fake_callback = mock.Mock(spec=lambda **kwargs: None)
        fake_callback.return_value = None

        fake.clean_ran.connect(Slot(fake_callback))
        fake.call_clean_run()

        fake_callback.assert_called_once_with()


@pytest.mark.parametrize("format", LogFormat)
@pytest.mark.parametrize("msg, kwargs, to_log, emitted", [
    ("zombie call_setup_run", dict(), False, None),
    ("useless start call_clean_run", dict(activate=True), False, None),
    (f"{Module.CONTEXT_EXIT_MESSAGE} call_clean_run", dict(activate=False), False, dict(levelno=LogLevel.DEBUG, pattern=r"Compute calls for [\w\.]+: \d+")),
    ("common message", dict(), True, None),
])
def test_Module_log_debug_message(format, msg, kwargs, to_log, emitted):
    handler = mock.MagicMock(level=LogLevel.DEBUG, log=mock.MagicMock())
    rec = logging.getLogRecordFactory()("log_test", LogLevel.INFO, __file__, 22, msg, (), None)
    for key, value in kwargs.items():
        setattr(rec, key, value)
    
    m = Module("dummy")

    assert m.log_debug_message(handler, rec, format) == to_log

    if emitted:
        handler.log.assert_called_once()
        args = handler.log.call_args[0]
        assert args[0] == emitted["levelno"]
        assert re.match(emitted["pattern"], args[1]) is not None
    else:
        handler.log.assert_not_called()



"""Mardown viewers for Module and Port."""
from cosapp.ports.port import ExtensiblePort
from cosapp.systems import System


def port_to_md(port: ExtensiblePort) -> str:
    """Returns the representation of this port variables in Markdown format.

    Parameters
    ----------
    port : ExtensiblePort
        Port to describe
        
    Returns
    -------
    str
        Markdown formatted representation
    """
    return "\n".join([
        f"- {value!r}" for value in port.get_details().values()
    ])


def system_to_md(system: System) -> str:
    """Returns the representation of a system in Markdown format.

    Parameters
    ----------
    system : System
        System to describe

    Returns
    -------
    str
        Markdown formatted representation
    """
    doc = list()
    if len(system.tags) > 0:
        doc.extend(["", f"**Tags**: {list(system.tags)!s}", ""])

    if len(system.children) > 0:
        doc.extend(["", "### Child components", ""])
        for name, child in system.children.items():
            doc.append(f"- `{name}`: {type(child).__name__}")

    if hasattr(system, 'residues') and len(system.residues) > 0:
        doc.extend(["", "### Residues", ""])
        doc.append(", ".join([f"`{key}`" for key in system.residues]))

    common_inputs = [System.INWARDS]
    common_outputs = [System.OUTWARDS]
    has_ports = len(system.inputs) + len(system.outputs) > len(common_inputs) + len(common_outputs)

    if has_ports:
        doc.extend(["", "### Ports", ""])

        def dump_port_data(header, port_dict, common_ports):
            if len(port_dict) <= len(common_ports):
                return
            doc.extend(["", f"#### {header.title()}", ""])
            for name, port in filter(lambda item: item[0] not in common_ports, port_dict.items()):
                doc.append(f"- `{name}`:")
                doc.extend([f"  {line}" for line in port_to_md(port).splitlines()])

        dump_port_data("inputs", system.inputs, common_inputs)
        dump_port_data("outputs", system.outputs, common_outputs)

    for name in common_inputs + common_outputs:
        if len(system[name]) > 0:
            doc.extend(["", f"#### {name.capitalize()}", ""])
            doc.append(port_to_md(system[name]))

    return "\n".join(doc)



import os
import unittest

def run_pytest():
    import pytest
    path = os.path.join(os.path.dirname(__file__), "..")
    return pytest.main([path, ])


class TestAll(unittest.TestCase):
    
    def test_pytest(self):
        self.assertEqual(run_pytest(), 0)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestAll('test_pytest'))
    return suite


if __name__ == "__main__":
    # Be sure to be on the top level directory of cosapp package

    # Optional dependency
    import coverage

    cover = coverage.Coverage()
    cover.start()
    result = run_pytest()
    cover.stop()
    cover.save()

    if result == 0:
        try:
            cover.report()
            cover.html_report()
        except coverage.misc.CoverageException as err:
            print(repr(err))
    else:
        exit(1)



from numbers import Number
from typing import Iterable, List, Tuple, Union

import numpy

from cosapp.core.eval_str import EvalString
from cosapp.core.numerics.basics import MathematicalProblem
from cosapp.core.numerics.boundary import Unknown
from cosapp.drivers.iterativecase import IterativeCase
from cosapp.utils.helpers import check_arg


class RunOptim(IterativeCase):
    """Driver running the model on its `System` owner and gathering optimization definition.

    Parameters
    ----------
    name : str
        Name of the driver
    owner : System, optional
        :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
    **kwargs : Dict[str, Any]
        Optional keywords arguments
    """

    __slots__ = ('constraints', )

    def __init__(self,
        name: str,
        owner: "Optional[cosapp.systems.System]" = None,
        **kwargs
    ) -> None:
        """Initialize a driver

        Parameters
        ----------
        name: str, optional
            Name of the `Module`
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
        **kwargs : Dict[str, Any]
            Optional keywords arguments
        """
        super().__init__(name, owner, **kwargs)
        # TODO we need to move this in an enhanced MathematicalProblem
        self.constraints = list()  # type: List[Dict]
            # desc="Constraints to be applied in the optimization problem"

    def set_objective(self, expression: str) -> None:
        """Set the scalar objective function to be minimized.

        Parameters
        ----------
        expression : str
            The objective expression to be minimized.
        """
        if self.owner is None:
            raise AttributeError("Owner System is needed to define an optimization.")

        check_arg(expression, "expression", str, lambda s: "==" not in s)

        self.design.residues.clear()  # Ensure only one objective is defined
        self.design.add_equation(expression + " == 0", name="f_objective", reference=1.)

    def add_unknown(self,
            name: Union[str, Iterable[Union[dict, str, Unknown]]],
            max_abs_step: Number = numpy.inf,
            max_rel_step: Number = numpy.inf,
            lower_bound: Number = -numpy.inf,
            upper_bound: Number = numpy.inf
    ) -> "MathematicalProblem":
        """Add unknown variables.

        You can set variable one by one or provide a list of dictionary to set multiple variable at once. The
        dictionary key are the arguments of this method.

        Parameters
        ----------
        name : str or Iterable of dictionary or str
            Name of the variable or list of variable to add
        max_rel_step : float, optional
            Maximal relative step by which the variable can be modified by the numerical solver; default numpy.inf
        max_abs_step : float, optional
            Maximal absolute step by which the variable can be modified by the numerical solver; default numpy.inf
        lower_bound : float, optional
            Lower bound on which the solver solution is saturated; default -numpy.inf
        upper_bound : float, optional
            Upper bound on which the solver solution is saturated; default numpy.inf

        Returns
        -------
        MathematicalProblem
            The modified MathematicalSystem
        """
        self.design.add_unknown(name, max_abs_step, max_rel_step, lower_bound, upper_bound)

    def add_constraints(self, expression: Union[str, List[Union[str, Tuple[str, bool]]]],
                        inequality: bool = True) -> None:
        """Add constraints to the optimization problem.

        Parameters
        ----------
        expression : str
            The expression defining the constraint
        inequality : bool
            If True, expression must be non-negative; else must be zero.
        """

        def add_constraint(expression: str, inequality: bool) -> None:
            check_arg(expression, 'expression', str)
            check_arg(inequality, 'inequality', bool)

            # Test that the expression can be evaluated
            EvalString(expression, self.owner).eval()

            self.constraints.append({
                'type': 'ineq' if inequality else 'eq',
                'formula': expression
            })

        if self.owner is None:
            raise AttributeError("Owner System is needed to define an optimization.")

        if isinstance(expression, str):
            add_constraint(expression, inequality)
        else:
            for args in expression:
                if isinstance(args, str):
                    add_constraint(args, inequality)
                else:
                    add_constraint(*args)

    def get_problem(self) -> MathematicalProblem:
        """Returns the full mathematical for the case.

        Returns
        -------
        MathematicalProblem
            The full mathematical problem to solve for the case
        """
        # TODO Go further to gather optimization unknowns defined in System hierarchy
        return self.design

    def setup_run(self):
        """Method called once before starting any simulation."""
        # Skip checking problem shape
        pass  # pragma: no cover



import pytest
import numpy
import scipy

from cosapp.systems import System
from cosapp.ports import Port
from cosapp.core.numerics.residues import Residue

# <codecell>

class BogusPort(Port):
    def setup(self):
        self.add_variable('q')

class FooBar(System):
    def setup(self):
        self.add_outward('z')
        self.add_input(BogusPort, 'in_', variables={'q': 5.})

class Bogus(System):
    def setup(self):
        self.add_property('g', 9.80665)
        self.add_property('NA', 6.02214076e23)
        self.add_inward('a', 2.0)
        self.add_inward('x', numpy.r_[0.1, -0.2, -3.14])
        self.add_inward('y', numpy.ones(3))
        self.add_outward('b', 0.5)

        self.add_child(FooBar('sub'))
        self.add_output(BogusPort, 'out', variables={'q': 0.5})

@pytest.fixture(scope="function")
def eval_context():
    """Returns a bogus system used only to provide a valid context"""
    return Bogus('bogus')

# <codecell>

@pytest.fixture(scope="function")
def ufunc_test_data():
    ufunc_data = {
        'arange': {'str': 'arange(0, 10, 2)', 'check_val': numpy.arange(0, 10, 2)},
        'array': {'str': 'array(x)',
                    'check_func': numpy.array,
                    'args': {'x': {'value': [-1., 0., 1.]}}},
        'asarray': {'str': 'asarray(x)',
                    'check_func': numpy.asarray,
                    'args': {'x': {'value': [-1., 0., 1.]}}},
        'concatenate': {
            'str': 'concatenate(x)',
            'check_func': numpy.concatenate,
            'args': {
                'x': {'value': (numpy.random.random(6), numpy.random.random(4))},
                },
            },
        'dot': {'str': 'dot(x, y)',
                'check_func': numpy.dot,
                'args': {
                    'x': {'value': numpy.random.random(6)},
                    'y': {'value': numpy.random.random(6)}}},
        'evaluate_residue': {'str': 'evaluate_residue(x, y)',
                    'check_func': Residue.evaluate_residue,
                    'args': {
                        'x': {'value': 25.},
                        'y': {'value': -12.}}},
        'fmax': {'str': 'fmax(x, y)',
                    'check_func': numpy.fmax,
                    'args': {
                        'x': {'value': numpy.random.random(6)},
                        'y': {'value': numpy.random.random(6)}}},
        'fmin': {'str': 'fmin(x, y)',
                    'check_func': numpy.fmin,
                    'args': {
                        'x': {'value': numpy.random.random(6)},
                        'y': {'value': numpy.random.random(6)}}},
        'inner': {'str': 'inner(x, y)',
                    'check_func': numpy.inner,
                    'args': {
                        'x': {'value': numpy.random.random(6)},
                        'y': {'value': numpy.random.random(6)}}},
        'isinf': {'str': 'isinf(x)',
                    'check_func': numpy.isinf,
                    'args': {
                        'x': {'value': [0, numpy.inf, 5.0]}}},
        'isnan': {'str': 'isnan(x)',
                    'check_func': numpy.isnan,
                    'args': {
                        'x': {'value': [0, numpy.nan, numpy.nan]}}},
        'kron': {'str': 'kron(x, y)',
                    'check_func': numpy.kron,
                    'args': {
                        'x': {'value': numpy.random.random(6)},
                        'y': {'value': numpy.random.random(6)}}},
        'linspace': {'str': 'linspace(0, 10, 50)',
                'check_val': numpy.linspace(0, 10, 50),
            },
        'matmul': {'str': 'matmul(x, y)',
                    'check_func': numpy.matmul,
                    'args': {
                        'x': {'value': numpy.random.random((3, 3))},
                        'y': {'value': numpy.random.random((3, 1))}}},
        'maximum': {'str': 'maximum(x, y)',
                    'check_func': numpy.maximum,
                    'args': {
                        'x': {'value': numpy.random.random(6)},
                        'y': {'value': numpy.random.random(6)}}},
        'minimum': {'str': 'minimum(x, y)',
                    'check_func': numpy.minimum,
                    'args': {
                        'x': {'value': numpy.random.random(6)},
                        'y': {'value': numpy.random.random(6)}}},
        'ones': {'str': 'ones(21)', 'check_val': numpy.ones(21)},
        'full': {
                'str': 'full((3, 7), 3.14159)',
                'check_val': numpy.full((3, 7), 3.14159),
            },
        'full_like': {'str': 'full_like(x, 3.14159)',
                    'check_val': numpy.full((3, 7), 3.14159),
                    'args': {'x': {'value': numpy.random.random((3, 7))}}},
        'cross': {'str': 'cross(x, y)',
                    'check_func': numpy.cross,
                    'args': {
                        'x': {'value': numpy.random.random(3)},
                        'y': {'value': numpy.random.random(3)}}},
        'outer': {'str': 'outer(x, y)',
                    'check_func': numpy.outer,
                    'args': {
                        'x': {'value': numpy.random.random(6)},
                        'y': {'value': numpy.random.random(6)}}},
        'power': {'str': 'power(x, y)',
                    'check_func': numpy.power,
                    'args': {
                        'x': {'value': numpy.random.random(6)},
                        'y': {'value': numpy.random.random(6) + 1.0}}},
        'residue_norm': {'str': 'residue_norm(x, y)',
                    'check_func': Residue.residue_norm,
                    'args': {
                        'x': {'value': 25.},
                        'y': {'value': -12.}}},
        'round': {'str': 'round(x, y)',
            'check_func': numpy.round,
            'args': {
                'x': {'value': numpy.random.random(6)},
                'y': {'value': 4},
                },
            },
        'tensordot': {'str': 'tensordot(x, y)',
                        'check_func': numpy.tensordot,
                        'args': {
                            'x': {'value': numpy.random.random((6,6))},
                            'y': {'value': numpy.random.random((6,6))}}},
        'zeros': {'str': 'zeros(21)', 'check_val': numpy.zeros(21)}
    }

    # Add base types
    for dtype in (
        numpy.int8, numpy.int16, numpy.int32, numpy.int64,
        numpy.uint8, numpy.uint16, numpy.uint32, numpy.uint64,
        numpy.float32, numpy.float64,
        numpy.complex64, numpy.complex128,
    ):
        name = dtype.__name__
        ufunc_data[name] = {
            'str': f'zeros(5, dtype={name})',
            'check_val': numpy.zeros(5, dtype=dtype)
        }

    def add_func(func, x=numpy.random.random(10), alias=None):
        name = alias or func.__name__
        ufunc_data[name] = {
            'str': f'{name}(x)',
            'check_func': func,
            'args': {'x': {'value': x}},
        }

    # Add simple ufuncs
    for func in (
        numpy.cos, numpy.sin, numpy.tan, numpy.arctan,
        numpy.cosh, numpy.sinh, numpy.tanh, numpy.arctanh,
        numpy.exp, numpy.expm1,
        numpy.log, numpy.log10, numpy.log1p,
        numpy.sqrt, numpy.cbrt,
        numpy.prod, numpy.sum,
        numpy.linalg.norm,
        scipy.special.factorial,
        scipy.special.erf, scipy.special.erfc,
    ):
        add_func(func, numpy.random.random(10))

    for func in (numpy.arccos, numpy.arcsin):
        add_func(func, numpy.random.random(6) - 0.5)

    for func in (numpy.arccosh, numpy.arcsinh):
        add_func(func, numpy.random.random(6) + 1.1)

    # Add common short names
    add_func(numpy.abs, alias='abs')
    add_func(numpy.arctan, alias='atan')
    add_func(numpy.arcsin, alias='asin', x=numpy.random.random(6) - 0.5)
    add_func(numpy.arccos, alias='acos', x=numpy.random.random(6) - 0.5)
    add_func(numpy.arctanh, alias='atanh')
    add_func(numpy.arcsinh, alias='asinh')
    add_func(numpy.arccosh, alias='acosh', x=numpy.random.random(6) + 1.1)

    # Add constants
    for name in ('e', 'pi', 'inf'):
        ufunc_data[name] = {'str': name, 'check_val': getattr(numpy, name)}

    return ufunc_data



"""
Ports package of CoSApp.
"""
# This is a pure trick to have a public API similar to a project
# Ports must be part of the core as Module depends on them.
from cosapp.ports.enum import PortType, Scope, Validity
from cosapp.ports.exceptions import ScopeError
from cosapp.ports.port import Port
from cosapp.ports.units import UnitError, add_offset_unit, add_unit, convert_units

__all__ = [
    "Port",
    "PortType",
    "Scope",
    "ScopeError",
    "Validity",
    "UnitError",
    "add_unit",
    "add_offset_unit",
    "convert_units",
]



import pytest
import numpy as np

from cosapp.drivers.time.scenario import Interpolator


@pytest.fixture(scope='function')
def data():
    return [[0, 0], [1, 1], [10, -17]]


@pytest.mark.parametrize("kind", Interpolator.Kind)
@pytest.mark.parametrize("data, expected", [
    ([[0, 1, 10], [0, 1, -17]], dict()),
    ([[0, 0], [1, 1]], dict(data=[[0, 1], [0, 1]])),
    ([[0, 0], [1, 1], [10, -17]], dict(data=[[0, 1, 10], [0, 1, -17]])),
    ([(0, 0), (1, 1), (10, -17)], dict(data=[[0, 1, 10], [0, 1, -17]])),
    ([[0], [1, 1], [10, -17]], dict(error=ValueError)),
    ([[0, 1, 2], [1, 2, 5, -17]], dict(error=ValueError)),
    ([[0, 0]], dict(error=ValueError)),
    ("foobar", dict(error=ValueError)),
    (dict(x=[0, 1], y=[1, 0.3]), dict(error=TypeError)),
])
def test_Interpolator__init__(data, kind, expected):
    error = expected.get('error', None)

    if error is None:
        interpolator = Interpolator(data, kind)
        assert interpolator.kind is kind
        assert np.array_equal(interpolator.data, expected.get('data', data))
    else:
        with pytest.raises(error, match=expected.get('match', None)):
            Interpolator(data, kind)


@pytest.mark.parametrize("kind, points, expected", [
    (Interpolator.Kind.Linear, [0, 0.5, 1, 2, 5.5], [0, 0.5, 1, -1, -8]),  # F(x) = x if x < 1 else 3 - 2 * x
    (Interpolator.Kind.CubicSpline, [0, 0.5, 1, 2, 5.5], [0, 0.575, 1, 1.4, -1.925]),
    (Interpolator.Kind.Pchip, [0, 0.5, 1, 2, 5.5], [0, 0.6625, 1, 0.84691358, -2.7125]),
])
def test_Interpolator__call__(kind, data, points, expected):
    interpolator = Interpolator(data, kind)
    assert interpolator(points) ==  pytest.approx(expected, rel=1e-8)


def test_Interpolator_kind(data):
    interpolator = Interpolator(data)
    points = [0, 0.5, 1, 2, 5.5]

    interpolator.kind = Interpolator.Kind.Linear
    assert interpolator.kind is Interpolator.Kind.Linear
    assert interpolator(points) ==  pytest.approx([0, 0.5, 1, -1, -8], rel=1e-15)

    interpolator.kind = Interpolator.Kind.CubicSpline
    assert interpolator.kind is Interpolator.Kind.CubicSpline
    assert interpolator(points) ==  pytest.approx([0, 0.575, 1, 1.4, -1.925], rel=1e-15)

    interpolator.kind = Interpolator.Kind.Pchip
    assert interpolator.kind is Interpolator.Kind.Pchip
    assert interpolator(points) ==  pytest.approx([0, 0.6625, 1, 0.84691358, -2.7125], rel=1e-8)



import numpy
from numbers import Number

from cosapp.drivers.time.interfaces import ExplicitTimeDriver
from cosapp.utils.helpers import check_arg


class RungeKutta(ExplicitTimeDriver):
    """
    Implementation of a few Runge-Kutta methods for time integration.
    https://en.wikipedia.org/wiki/List_of_Runge%E2%80%93Kutta_methods
    """

    __slots__ = ('__fracs', '__coefs', '__buffer', '__init_name')

    def __init__(self, name="", owner: "Optional[cosapp.systems.System]" = None, order: int = 2, **options):
        """Initialization of the driver

        Parameters
        ----------
        owner: System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belongs; default None
        order: int, optional
            Order of the numerical scheme
        **options: Dict[str, Any]
            Optional keywords arguments; may contain time step and interval, with keys `dt` and `time_interval`
        """
        check_arg(name, "name", str)
        super().__init__("temp", owner, **options)
        self.__init_name = self._name_check(name) if len(name) > 0 else name
        self.__fracs = None
        self.__coefs = None
        self.__buffer = None
        self.order = order

    @property
    def algo_info(self) -> str:
        """str: Short description of driver's algorithm"""
        return f"Explicit order-{self.order} Runge-Kutta time driver"

    @property
    def name(self) -> str:
        return self.__init_name or self.algo_info

    @property
    def order(self) -> int:
        """Numerical order of the time integration scheme"""
        return len(self.__coefs)

    @order.setter
    def order(self, order: int) -> None:
        check_arg(order, 'order', int, lambda n: 2 <= n <= 4)
        if order == 2:
            # Ralston's second-order scheme
            self.__fracs = numpy.r_[2 / 3]
            self.__coefs = numpy.r_[0.25, 0.75]
        elif order == 3:
            # Heun's third-oder scheme
            self.__fracs = numpy.r_[1, 2] / 3
            self.__coefs = numpy.r_[0.25, 0, 0.75]
        elif order == 4:
            # Runge-Kutta fourth-order scheme
            self.__fracs = numpy.r_[0.5, 0.5, 1]
            self.__coefs = numpy.r_[1, 2, 2, 1] / 6

    def _precompute(self) -> None:
        super()._precompute()
        # Create memory buffer to store nstages intermediate values of dx/dt, and x at t = tn
        nstages = len(self.__coefs)
        self.__buffer = { name : [x.value] * (nstages + 1) for name, x in self._transients.items() }

    def _update_transients(self, dt: Number) -> None:
        """
        Time integration of transient variables over time step `dt` by Runge-Kutta scheme.
        """
        transients = self._transients
        buffer = self.__buffer
        steps = self.__fracs * dt
        weights = self.__coefs * dt
        tn = self.time

        # Store unknown at t = tn
        for name, x in transients.items():
            buffer[name][-1] = x.value
        
        # Compute sub-step stages
        for stage, dts in enumerate(steps):
            for name, x in transients.items():
                buffer[name][stage] = dx_dt = x.d_dt
                x.value = buffer[name][-1] + dx_dt * dts
            self._set_time(tn + dts)

        # Compute weighted solution at time tn + dt
        for name, x in transients.items():
            buffer[name][-2] = x.d_dt
            new_x = buffer[name][-1]
            for stage, weight in enumerate(weights):
                new_x += weight * buffer[name][stage]
            x.value = new_x



import logging
from collections import OrderedDict
from typing import Any, Dict, Optional, Union

import numpy

from cosapp.core.eval_str import AssignString
from cosapp.core.numerics.basics import MathematicalProblem
from cosapp.core.numerics.boundary import Boundary
from cosapp.drivers.iterativecase import IterativeCase
from cosapp.ports.enum import PortType
from cosapp.ports.port import ExtensiblePort
from cosapp.ports.variable import ArrayIndices
from cosapp.systems import System
from cosapp.utils.helpers import check_arg

logger = logging.getLogger(__name__)


class RunSingleCase(IterativeCase):
    """Set new boundary conditions and equations on the system.

    By default, it has a :py:class:`~cosapp.drivers.runonce.RunOnce` driver as child to run the system.

    Attributes
    ----------
    case_values : List[AssignString]
        List of requested variable assignments to set up the case
    initial_values : Dict[str, Tuple[Any, Optional[numpy.ndarray]]]
        List of variables to set with the values to set and associated indices selection

    Parameters
    ----------
    name : str
        Name of the driver
    owner : System, optional
        :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
    **kwargs : Any
        Keyword arguments will be used to set driver options
    """

    __slots__ = ('__case_values', 'offdesign', 'problem')

    def __init__(self,
        name: str,
        owner: Optional[System] = None,
        **kwargs
        ) -> None:
        """Initialize a driver

        Parameters
        ----------
        name: str, optional
            Name of the `Module`
        owner : System, optional
            :py:class:`~cosapp.systems.system.System` to which this driver belong; default None
        **kwargs : Dict[str, Any]
            Optional keywords arguments
        """
        super().__init__(name, owner, **kwargs)
        self.__case_values = []  # type: List[AssignString]
            # desc="List of assignments 'lhs <- rhs' to perform in the present case.")
        self.owner = owner
        self.offdesign = MathematicalProblem(self.name + "- offdesign", self.owner)  # type: MathematicalProblem
            # desc="Additional mathematical problem to solve for on this case only.")
        self.problem = None  # type: Optional[MathematicalProblem]
            # desc='Full mathematical problem to be solved on this case.'

    def setup_run(self):
        """Method called once before starting any simulation."""
        super().setup_run()
        
        full_name = lambda name: self.name + name.join('[]')

        def raise_ValueError(kind, name):
            raise ValueError(
                "{} {!r} is defined as design and offdesign {} in driver {!r}".format(
                    kind.capitalize(), name, kind.lower(), self.name))

        self.problem = MathematicalProblem(self.name, self.owner)
        # Add design equations
        for name, unknown in self.design.unknowns.items():
            self.problem.unknowns[name] = unknown
        for name, residue in self.design.residues.items():
            self.problem.residues[full_name(name)] = residue
        # Add off-design equations
        for name, unknown in self.offdesign.unknowns.items():
            if name in self.problem.unknowns:
                raise_ValueError("variable", name)
            self.problem.unknowns[full_name(name)] = unknown
        for name, residue in self.offdesign.residues.items():
            fullname = full_name(name)
            if fullname in self.problem.residues:
                raise_ValueError("equation", name)
            self.problem.residues[fullname] = residue
        # Get common off-design problem to be solved on each case
        common_system = self.owner.get_unsolved_problem()
        # Add common off-design equations taken into account switch in frozen status
        for name, unknown in common_system.unknowns.items():
            if name in self.problem.unknowns:
                raise_ValueError("variable", name)
            # Common unknowns must be duplicated to avoid modification by one point to the others
            self.problem.unknowns[full_name(name)] = unknown.copy()
        for name, residue in common_system.residues.items():
            fullname = full_name(name)
            if fullname in self.problem.residues:
                raise_ValueError("equation", name)
            # Common residues must be duplicated to avoid modification by one point to the others
            self.problem.residues[fullname] = residue.copy()

    def _precompute(self) -> None:
        """Actions to carry out before the :py:meth:`~cosapp.drivers.runonce.RunOnce.compute` method call.

        It sets the boundary conditions and changes variable status.
        """
        super()._precompute()

        # Set the boundary conditions
        for assignment in self.case_values:
            value, changed = assignment.exec()
            if changed:
                self.owner.set_dirty(PortType.IN)

        # Set the offdesign variables
        for name, unknown in self.get_problem().unknowns.items():
            if name not in self.design.unknowns and not numpy.array_equal(unknown.value, unknown.default_value):
                unknown.set_to_default()

    def clean_run(self):
        """Method called once after any simulation."""
        self.problem = None

    @IterativeCase.owner.setter
    def owner(self, value: Optional[System]) -> None:
        # Trick to call super setter (see: https://bugs.python.org/issue14965)
        if self.owner is not value:
            if self.owner is not None:
                logger.warning(
                    f"System owner of Driver {self.name!r} has changed. Design and offdesign equations have been cleared."
                )
            self.offdesign = MathematicalProblem(self.offdesign.name, value)
        super(RunSingleCase, RunSingleCase).owner.__set__(self, value)

    def get_problem(self) -> MathematicalProblem:
        """Returns the full mathematical for the case.

        Returns
        -------
        MathematicalProblem
            The full mathematical problem to solve for the case
        """
        if self.problem is None:
            logger.warning("RunSingleCase.get_problem called with no prior call to RunSingleCase.setup_run.")
            return MathematicalProblem(self.name, self.owner)
        else:
            return self.problem

    def set_values(self, modifications: Dict[str, Any]) -> None:
        """Enter the set of variables defining the case, from a dictionary of the kind {'variable1': value1, ...}
        Note: will erase all previously defined values. Use 'add_values' to append new case values.

        The variable can be contextual `child1.port2.var`. The only rule is that it should belong to
        the owner `System` of this driver or any of its descendants.

        Parameters
        ----------
        modifications : Dict[str, Any]
            Dictionary of (variable name, value)

        Examples
        --------
        >>> driver.set_values({'myvar': 42, 'port.dummy': 'banana'})
        """
        self.clear_values()
        self.add_values(modifications)

    def add_values(self, modifications: Dict[str, Any]) -> None:
        """Add a set of variables to the list of case values, from a dictionary of the kind {'variable1': value1, ...}

        The variable can be contextual `child1.port2.var`. The only rule is that it should belong to
        the owner `System` of this driver or any of its descendants.

        Parameters
        ----------
        modifications : Dict[str, Any]
            Dictionary of (variable name, value)

        Examples
        --------
        >>> driver.add_values({'myvar': 42, 'port.dummy': 'banana'})
        """
        check_arg(modifications, 'modifications', dict)

        for variable, value in modifications.items():
            self.add_value(variable, value)

    def add_value(self, variable: str, value: Any) -> None:
        """Add a single variable to list of case values.

        The variable can be contextual `child1.port2.var`. The only rule is that it should belong to
        the owner `System` of this driver or any of its descendants.

        Parameters
        ----------
        variable : str
            Name of the variable
        value : Any
            Value to be used.

        Examples
        --------
        >>> driver.add_value('myvar', 42)
        """
        if self.owner is None:
            raise AttributeError(
                f"Driver {self.name!r} must be attached to a System to set case values."
            )
        else:
            Boundary.parse(self.owner, variable)  # checks that variable is valid
            self.__case_values.append(AssignString(variable, value, self.owner))

    def clear_values(self):
        self.__case_values.clear()

    @property
    def case_values(self):
        return self.__case_values



import numpy
import collections
import pickle
import pandas
import logging
import warnings
from numbers import Number
from collections import OrderedDict
from typing import Any, Dict, Iterable, NamedTuple, NoReturn, Union, Optional, List

from cosapp.utils.helpers import check_arg
from cosapp.utils.logging import LogLevel
from cosapp.utils.find_variables import find_variables, make_wishlist, natural_varname
from cosapp.ports.enum import PortType


logger = logging.getLogger(__name__)


class SystemSurrogateState(NamedTuple):
    """Named tuple containing internal state data of a SystemSurrogate object"""
    doe_in: pandas.DataFrame
    doe_out: Union[pandas.DataFrame, OrderedDict]
    model: Any
    doe_out_sizes: OrderedDict


class SystemSurrogate:
    """
    Class representing a surrogate model trained from the outputs of an owner system.
    Once the surrogate is created, it supersedes (whenever activated) the behaviour of
    `System.compute()`. 
    """
    __slots__ = ('__owner', '__state', '__need_doe')

    def __init__(self,
        owner: "cosapp.systems.System",
        data_in: Union[pandas.DataFrame, Dict[str, Any]],
        model: type,
        data_out: Optional[Union[pandas.DataFrame, Dict[str, Any]]] = None,
        postsynch: Union[str, List[str]] = '*',
        *args, **kwargs
    ):
        # CHECK ARGS
        from cosapp.systems import System
        check_arg(owner, 'owner', System)
        check_arg(data_in, 'data_in', (pandas.DataFrame, dict))
        check_arg(data_out, 'data_out', (pandas.DataFrame, dict, type(None)))
        if model is not None:
            check_arg(model, 'model', type,
                lambda m: all(hasattr(m, attr) for attr in ('train', 'predict'))
            )
        postsynch = make_wishlist(postsynch, 'postsynch')

        if owner is not None:
            logger.debug(f"Initialize surrogateclass for System {owner.name!r}, data_in as {data_in} and model as {model}")

        # BUILD INTERN DATA
        self.__owner = owner
        self.__need_doe = data_out is None
        
        doe_in = pandas.DataFrame.from_dict(data_in) if isinstance(data_in, dict) else data_in
        doe_out = self.__init_doe_out(data_out, postsynch)
        model_obj = None if model is None else model(*args, **kwargs)
        self.__state = SystemSurrogateState(doe_in, doe_out, model_obj, OrderedDict())
        
        # TRAINING
        empty = any(obj is None for obj in (owner, model_obj)) or len(data_in) == 0
        if not empty:
            self.__check_unknowns_and_transients()
            self.__prepare_and_train()

        logger.debug(f"surrogateclass initialized")

    @property
    def state(self)-> SystemSurrogateState:
        return self.__state

    @property
    def owner(self)-> "cosapp.systems.System":
        return self.__owner

    @property
    def synched_outputs(self) -> List[str]:
        """List[str]: list of synchronized output variable names"""
        return list(self.__state.doe_out_sizes.keys())

    def __init_doe_out(self, data_out, postsynch) -> "OrderedDict[str, List]":
        doe_out = OrderedDict()
        if isinstance(data_out, pandas.DataFrame):
            doe_out = data_out.to_dict(into=OrderedDict)
        elif data_out is not None:
            doe_out = OrderedDict(data_out)
        elif self.owner is not None:
            watched = self.__get_owner_connections()
            varlist = watched.keys()
            if '*' not in postsynch:
                owner = self.owner
                def writeable(var) -> bool:
                    try:
                        owner[var] = owner[var]
                    except AttributeError:
                        return False
                    else:
                        return True
                varlist = find_variables(owner,
                        includes=postsynch,
                        excludes=None,
                        inputs=False,
                    )
                varlist = set(filter(writeable, varlist))
                # Make sure varlist contains at least owner system outputs
                for port, output in owner.outputs.items():
                    varlist |= set(natural_varname(f"{port}.{var}") for var in output)
            doe_out = OrderedDict((var, []) for var in varlist)
        return doe_out

    def __get_doe_out_sizes(self) -> None:
        state = self.__state
        owner = self.__owner
        doe_out_sizes = OrderedDict.fromkeys(state.doe_out.keys(), None)
        pos = 0
        for var in doe_out_sizes:
            size = 0
            value = owner[var]
            if isinstance(value, Number):
                size = 1
            elif isinstance(value, numpy.ndarray):
                size = value.size
            else:
                raise TypeError(
                    f"Unsupported data type {type(value).__name__} for {owner.name}.{var}"
                )
            doe_out_sizes[var] = (pos, size)
            pos += size
        self.__state = SystemSurrogateState(
            state.doe_in,
            state.doe_out,
            state.model,
            doe_out_sizes,
        )

    def __format_outputs(self) -> numpy.ndarray:
        logger.debug(f"Reshaping outputs")
        doe_length = len(self.__state.doe_in)
        doe_out = self.__state.doe_out
        reshaped_outputs = []
        for k in range(doe_length):
            reshaped_outputs.append(
                list(value[k] for value in doe_out.values())
            )
        reshaped_outputs = numpy.asarray(list(list(flatten(el)) for el in reshaped_outputs))
        # logger.debug(f"Reshaped outputs are now numpy.ndarray and are: {reshaped_outputs}")
        return reshaped_outputs

    def __format_inputs(self) -> numpy.ndarray:
        logger.debug(f"Reshaping input data")
        res = []
        for couple_values in self.__state.doe_in.values:
            res.append(list(flatten(couple_values)))
        return numpy.asarray(res)

    def __set_and_execute(self) -> None:
        logger.debug(f"Setting and executing in order to build data for training")
        owner = self.__owner
        doe_out = self.__state.doe_out
        for i, row in self.state.doe_in.iterrows():
            logger.log(LogLevel.FULL_DEBUG, f"Setting {owner.name!r} input values (DOE row #{i})")
            for var, value in row.items():
                owner[var] = value
            owner.run_drivers()
            # add tracked output data to doe_out
            for var in doe_out:
                doe_out[var].append(owner[var])

    def add_data(self, newdoe: pandas.DataFrame) -> NoReturn:
        # TODO: Should merge input lists when matching names (not working yet)
        # self.__state = SystemSurrogateState(
        #     self.__state.doe_in,
        #     self.__state.doe_out,
        #     self.__state.model,
        #     OrderedDict(),
        # )
        raise NotImplementedError()

    def __prepare_and_train(self) -> None:
        logger.debug(f"Preparing and training function")
        if len(self.state.doe_out) == 0:
            raise ValueError(
                f"Cannot train surrogate model: no output found in System {self.owner.name!r}"
            )
        self.__get_doe_out_sizes()
        if self.__need_doe:
            self.__set_and_execute()
        x = self.__format_inputs()
        y = self.__format_outputs()
        self.__train_model(x, y)

    def __train_model(self, x, y) -> None:
        model = self.__state.model
        logger.debug(f"Training model {model} with \n\tX = {x}\n\tY = {y}")
        model.train(x, y)
        logger.debug(f"Model {model} trained")

    def predict(self, x):
        return self.__state.model.predict(x).reshape(1, -1)[0]

    def compute(self) -> None:
        logger.debug(f"Start meta_compute() instead of {self.owner.name}.compute().")
        inputs = self.__get_owner_inputs()
        outputs = self.predict(inputs)
        self.__set_owner_outputs(outputs)
        logger.debug(f"meta_compute() done")

    def __get_owner_inputs(self) -> numpy.ndarray:
        owner = self.owner
        logger.debug(f"Getting current inputs of system {owner.name}")
        inputs = list(flatten(owner[var] for var in self.__state.doe_in.columns))
        logger.debug(f"Detected input values: {inputs}")
        return numpy.asarray(inputs)

    def __set_owner_outputs(self, np_outputs: numpy.ndarray) -> None:
        owner = self.__owner
        logger.log(LogLevel.FULL_DEBUG, f"Post-synchronize outputs of {owner.name}")
        for var, (pos, size) in self.__state.doe_out_sizes.items():
            if size == 1:
                owner[var] = np_outputs[pos]
            elif size > 1:
                owner[var].ravel()[:] = np_outputs[pos : pos + size]

    def dump(self, filename: str) -> None:
        """Dump current state to file"""
        logger.debug(f"Dumping metamodel in {filename}")
        with open(filename, 'wb') as fp:
            pickle.dump(self.__state, fp)

    @classmethod
    def load(cls, owner: "cosapp.systems.System", filename: str) -> "SystemSurrogate":
        """Load system surrogate from file"""
        logger.debug(f"Loading metamodel from {filename}")
        from cosapp.systems import System
        check_arg(owner, 'owner', System)
        with open(filename, 'rb') as fp:
            state = pickle.load(fp)
        if not isinstance(state, SystemSurrogateState):
            raise TypeError("cannot load object")
        obj = cls(owner, data_in={}, model=None)
        obj.__state = state
        return obj

    def __check_unknowns_and_transients(self) -> None:
        state = self.state
        unknowns = get_unknows_transients(self.__owner)
        unsolvable_unknows = set(unknowns).difference(state.doe_out.keys())
        trained_unknowns = unsolvable_unknows.issubset(state.doe_in.keys())
        if not trained_unknowns:
            message = "The following unknowns/transients are not part of the training set; "\
                "future attempts to compute them with a driver may fail:"
            for name in unknowns.difference(state.doe_in.keys()):
                message += f"\n\t- {name}"
            warnings.warn(message)

    def __get_owner_connections(self) -> "OrderedDict[str, list]":
        return get_dependant_connections(self.__owner)


def flatten(iterable: Iterable) -> Iterable:
    for elem in iterable:
        if isinstance(elem, collections.abc.Iterable) and not isinstance(elem, str): 
            yield from flatten(elem)
        else:
            yield elem


def get_dependant_connections(system, head_system=None) -> Dict[str, list]:
    """
    This function returns a dictionnary mapping variable names to a port direction.
    Keys are absolute paths to connected inputs and all outputs.
    Values are owner port direction.
    """
    logger.debug(f"Starting recursive getter of connected inputs and all outputs on {system.name}")
    result = dict()
    prefix = ""

    if head_system is None:
        head_system = system
    elif system is not head_system:
        prefix = f"{head_system.get_path_to_child(system)}."
        for list_connectors in system.parent._connectors.values():
            for connector in list_connectors:
                sink = connector.sink
                mapping = connector.variable_mapping
                logger.debug(f"Detecting connector {connector} with sink {sink.name!r}")
                if sink in system.inputs.values():
                    for portvariable in mapping:
                        key = f"{prefix}{sink.name}.{portvariable}"
                        result[key] = PortType.IN
                        logger.debug(f"Add {key} to list of connected inputs")

    for output in system.outputs.values():
        logger.debug(f"Checking output {output}.")
        for portvariable in output:
            key = f"{prefix}{output.name}.{portvariable}"
            logger.debug(f"Add {key} to list of outputs")
            result[key] = PortType.OUT
    if system.is_standalone():
        unknowns = system.get_unsolved_problem().unknowns
        for unknown in unknowns:
            key = f"{prefix}{unknown}"
            result[key] = PortType.IN
            logger.debug(f"Add {key} to list of unknowns")
    for child in system.children.values():
        logger.debug(f"Targeted child of recursive getter of connected inputs and all outputs is {child}")
        child_dict_path_str = get_dependant_connections(child, head_system)
        result.update(child_dict_path_str)
    return result


def get_unknows_transients(system: "cosapp.systems.System") -> set:
    res = set()
    problem = system.get_unsolved_problem()
    for collection in (problem.unknowns, problem.transients):
        res |= set(collection)
    return res



"""Export a System as FMU."""
import itertools
import logging
import os
import shutil
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Dict, Iterable, List, NamedTuple, Optional, Union
from collections.abc import Collection

import numpy

from cosapp.core import __version__ as cosapp_version
from cosapp.drivers import RunSingleCase
from cosapp.drivers.abstractsolver import AbstractSolver
from cosapp.ports.port import ExtensiblePort
from cosapp.systems import System
from cosapp.utils.helpers import check_arg

try:
    from pythonfmu import Fmi2Causality, Fmi2Variability, FmuBuilder as PyFmuBuilder
except ImportError:

    class Fmi2Causality(Enum):
        parameter = 0
        calculatedParameter = 1
        input = 2
        output = 3
        local = 4

    class Fmi2Variability(Enum):
        constant = 0
        fixed = 1
        tunable = 2
        discrete = 3
        continuous = 4

    PyFmuBuilder = None


logger = logging.getLogger(__name__)

DECK_FILE_NAME = "system.json"
THIS_YEAR = datetime.now().year


class TimeIntegrator(Enum):
    """Available time integrator for FMU export."""

    EulerExplicit = "Euler explicit"
    RK2 = "Runge-Kutta 2"
    RK3 = "Runge-Kutta 3"
    RK4 = "Runge-Kutta 4"

    @property
    def driver(self) -> str:
        if self.name.startswith("RK"):
            return "RungeKutta"
        else:
            return self.name

    @property
    def options(self) -> Dict[str, str]:
        if self.name.startswith("RK"):
            return dict(order=int(self.name[2:]))
        else:
            return dict()


class VariableType(Enum):
    Boolean = "bool"
    Integer = "int"
    Real = "float"
    String = "bytes"


class Variable(NamedTuple):
    """Variable attributes"""

    name: str
    pytype: str  # Python type
    fmutype: str  # FMU type
    causality: Fmi2Causality
    variability: Optional[Fmi2Variability] = None


class FmuBuilder:
    """CoSApp FMU builder"""

    @staticmethod
    def _add_variables(
        vars: Dict[str, Any],
        causality: Fmi2Causality,
        variability: Optional[Fmi2Variability] = None,
    ) -> List[Variable]:
        """Transforms the to-included variable dictionary in 
        a list of Variable object for injection in the Jinja2 template.
        
        Parameters
        ----------
        vars : Dict[str, Any]
            Dictionary of the to-be-included variables
        causality : Fmi2Causality
            Causality of the variables
        variablility : Fmi2Variablity or None
            Variability of the variables; default not specified

        Returns
        -------
        List[Variable]
            List of Variable object to be injected in the template.
        """
        l_vars = list()
        for name, value in vars.items():
            try:
                dtype = FmuBuilder._get_variable_type(value)
            except TypeError as e:
                raise TypeError(f"{e!s} for variable {name}")

            l_vars.append(
                Variable(
                    name,
                    dtype.value,
                    dtype.name,
                    causality.name,
                    getattr(variability, "name", None),
                )
            )

        return l_vars

    @staticmethod
    def _get_default_value(names: Iterable[str], system: System) -> Dict[str, Any]:
        """Read the default value of the variable in the system.
        
        Parameters
        ----------
        names: Iterable[str]
            Variable to look for
        system: System
            CoSApp system in from which the values need to be taken

        Returns
        -------
        Dict[str, Any] : Dictionary of the variable name (key) with their value (value)
        """
        values = dict()
        for name in names:
            try:
                values[name] = eval(f"master.{name}", {"master": system})
            except (AttributeError, IndexError, KeyError):
                raise ValueError(
                    f"Variable {name} not found in the system '{system.name}'."
                )

        return values

    @staticmethod
    def _get_default_variables(
        ports: Dict[str, ExtensiblePort], to_skip: Iterable[str]
    ) -> Dict[str, Any]:
        """Get the default variable list from a port list.

        Variable names will be `port`.`variable_name` except for
        inwards and outwards (it will be `variable_name` directly).

        Only variables of type supported by FMI are listed.

        Parameters
        ----------
        ports : Dict[str, ExtensiblePort]
            List of port to extract variables from
        to_skip : Iterable[str]
            List of variable names to ignore

        Returns
        -------
        Dict[str, Any]
            Return a dictionary of the variables to be included with their
            default value.
        """
        vars = dict()
        for port_name, port in ports.items():
            for name in port:
                if port_name not in (System.INWARDS, System.OUTWARDS):
                    full_name = f"{port_name}.{name}"
                else:
                    full_name = name
                if full_name in to_skip:
                    continue

                value = port[name]
                try:
                    FmuBuilder._get_variable_type(value)
                except TypeError:
                    logger.debug(
                        f"Variable '{full_name}' has unsupported type for FMI."
                    )
                else:
                    vars[full_name] = value
        return vars

    @staticmethod
    def _get_documentation_folder(dest: Path) -> Path:
        """Path: The subfolder containing the documentation files."""
        doc_folder = dest / "documentation"
        doc_folder.mkdir(parents=True, exist_ok=True)
        return doc_folder

    @staticmethod
    def _get_project_folder(dest: Path) -> Path:
        """Path: The subfolder containing the project files."""
        project_folder = dest / "project_files"
        project_folder.mkdir(parents=True, exist_ok=True)
        return project_folder

    @staticmethod
    def _get_script_file(dest: Path, system: System, suffix: str) -> Path:
        """Path: Facade module file path."""
        class_name = type(system).__name__
        path = FmuBuilder._get_project_folder(dest)
        filename = path / f"{class_name.lower()}{suffix}.py"
        filename.parent.mkdir(parents=True, exist_ok=True)
        return filename

    @staticmethod
    def _get_variable_type(value: Any) -> VariableType:
        """Get the FMI variable type of a value.
        
        Parameters
        ----------
        value : Any
            Value to test
        
        Returns
        -------
        VariableType
            Type of the variable

        Raises
        ------
        TypeError
            For unsupported variable types.
        """

        if isinstance(value, (bool, numpy.bool_)):
            dtype = VariableType.Boolean
        elif isinstance(value, (int, numpy.int_)):
            dtype = VariableType.Integer
        elif isinstance(value, (float, numpy.float_)):
            dtype = VariableType.Real
        elif isinstance(value, (str, numpy.str_)):
            dtype = VariableType.String
        elif isinstance(value, numpy.ndarray):
            if value.ndim != 0:
                raise TypeError("Unsupported numpy array")
            if numpy.issubdtype(value.dtype, numpy.dtype(bool)):
                dtype = VariableType.Boolean
            elif numpy.issubdtype(value.dtype, numpy.dtype(int)):
                dtype = VariableType.Integer
            elif numpy.issubdtype(value.dtype, numpy.dtype(float)):
                dtype = VariableType.Real
            elif numpy.issubdtype(value.dtype, numpy.dtype(str)):
                dtype = VariableType.String
            else:
                raise TypeError(f"Unsupported numpy dtype {value.dtype}")
        else:
            dtype = type(value)
            raise TypeError(f"Unsupported type {dtype.__qualname__}")

        return dtype

    @staticmethod
    def generate_fmu_facade(
        system: System,
        inputs: Iterable[str] = None,
        parameters: Iterable[str] = None,
        outputs: Iterable[str] = None,
        locals: Iterable[str] = None,
        time_integrator: Union[TimeIntegrator, str] = TimeIntegrator.RK4,
        non_linear_solver: Optional[AbstractSolver] = None,
        dest: Union[Path, str] = os.curdir,
        python_env: Union[Path, str, None] = None,
        version: Optional[str] = None,
        author: Optional[str] = None,
        description: Optional[str] = None,
        copyright: Optional[str] = "",
        license: Optional[str] = "",
        fmu_name_suffix: str = "",
    ) -> Path:
        """Export a system as CoSimulation FMU respecting FMI 2.0.

        Inputs, parameters, outputs and locals are by default taken from inputs,
        inwards, outputs and outwards variables of the CoSApp system, respectively. If you wish
        to set an empty list, set the corresponding argument with a empty dictionary.

        You can add a mathematical problem to the system by passing the proper
        non_linear_solver. If it contains a RunSingleCase, its equations will be
        passed to the FMU. 
        
        Parameters
        ----------
        system : System
            System to be exported
        inputs : Iterable[str], optional
            List of input variables with initial values; default None
        parameters : Iterable[str], optional
            List of parameter variables with initial values; default None
        outputs : Iterable[str], optional
            List of output variables with initial values; default None
        locals : Iterable[str], optional
            List of local variables with initial values; default None
        time_integrator : TimeIntegrator, optional
            Time integrator algorithm; default Runge-Kutta 4th order
        non_linear_solver : AbstractSolver, optional
            Non linear Driver to use for solving the system at a given instant
        dest : str or Path, optional
            Destination folder; default current directory
        python_env : str or Path, optional
            File listing the python dependency; default None
        version : str, optional
            FMU version; default None
        author : str, optional
            FMU author; default None
        description : str, optional
            System description; default None
        license : str, optional
            FMU license; default ""
        copyright : str, optional
            FMU copyright; default ""
        fmu_name_suffix : str, optional
            FMU name; default ""

        Returns
        -------
        pathlib.Path
            Folder path containing FMU facade files
        """
        # Optional dependencies
        try:
            from jinja2 import Environment, PackageLoader
        except ImportError:
            raise ImportError("jinja2 needs to be installed to export a System as FMU.")

        # Check the arguments
        nonetype = type(None)
        check_arg(system, "system", System)
        check_arg(inputs, "inputs", (Collection, nonetype))
        check_arg(parameters, "parameters", (Collection, nonetype))
        check_arg(outputs, "outputs", (Collection, nonetype))
        check_arg(locals, "locals", (Collection, nonetype))
        check_arg(time_integrator, "time_integrator", (TimeIntegrator, str))
        check_arg(non_linear_solver, "non_linear_solver", (AbstractSolver, nonetype))
        check_arg(dest, "dest", (Path, str))
        check_arg(python_env, "python_env", (Path, str, nonetype))
        local_vars = dict()
        options = {
            "version": version,
            "author": author,
            "description": description,
            "copyright": copyright,
            "license": license,
        }
        for name, value in options.items():
            if value is not None:
                check_arg(value, name, str)
                local_vars[name] = value

        # Convert argument
        time_integrator = TimeIntegrator(time_integrator)
        temp_dest = Path(dest)

        # Generate the parameters
        system_type = type(system)
        params = {
            "module_name": system_type.__module__,
            "class_name": system_type.__qualname__,
            "driver": time_integrator.driver,
            "time_options": time_integrator.options,
            # TODO 'nl_options': dict(),
            "variables": list(),
            "class_attrs": dict(),
            "system_file": DECK_FILE_NAME,
        }

        for name, value in local_vars.items():
            params["class_attrs"][name] = value

        # TODO FMI supports to set for CoSimulation the DefaultExperiment.stepSize
        # that defines the preferred communicationStepSize

        user_list = set(
            itertools.chain(inputs or [], parameters or [], outputs or [], locals or [])
        )

        if inputs is None:
            inward_port = system.inputs.pop(System.INWARDS)
            inputs = FmuBuilder._get_default_variables(system.inputs, user_list)
            system.inputs[System.INWARDS] = inward_port  # Restore popped port
        else:
            inputs = FmuBuilder._get_default_value(inputs, system)
        if parameters is None:
            parameters = FmuBuilder._get_default_variables(
                {System.INWARDS: system.inputs[System.INWARDS]}, user_list
            )
        else:
            parameters = FmuBuilder._get_default_value(parameters, system)

        if outputs is None:
            outward_port = system.outputs.pop(System.OUTWARDS)
            outputs = FmuBuilder._get_default_variables(system.outputs, user_list)
            system.outputs[System.OUTWARDS] = outward_port
        else:
            outputs = FmuBuilder._get_default_value(outputs, system)

        if locals is None:
            locals = FmuBuilder._get_default_variables(
                {System.OUTWARDS: system.outputs[System.OUTWARDS]}, user_list
            )
        else:
            locals = FmuBuilder._get_default_value(locals, system)

        params["variables"].extend(
            FmuBuilder._add_variables(inputs, Fmi2Causality.input)
        )
        params["variables"].extend(
            FmuBuilder._add_variables(
                parameters, Fmi2Causality.parameter, Fmi2Variability.tunable
            )
        )
        params["variables"].extend(
            FmuBuilder._add_variables(outputs, Fmi2Causality.output)
        )
        params["variables"].extend(
            FmuBuilder._add_variables(locals, Fmi2Causality.local)
        )

        if non_linear_solver is not None:
            children_driver = list(non_linear_solver.children.values())
            if len(children_driver) > 1:
                raise ValueError("The non-linear solver can only have one child.")
            runsinglecase = children_driver[0]
            if not isinstance(runsinglecase, RunSingleCase):
                raise TypeError(
                    "The non-linear solver child must be of type RunSingleCase."
                )

            problem = runsinglecase.offdesign.copy()
            problem.extend(runsinglecase.design)
            params["problem"] = problem.to_dict()

        env = Environment(
            loader=PackageLoader("cosapp.tools", "templates"),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        template = env.get_template("pythonfmu.j2")

        rendered_script = template.render(params)

        filename = FmuBuilder._get_script_file(temp_dest, system, fmu_name_suffix)
        with open(filename, mode="w", encoding="utf-8") as f:
            f.write(rendered_script)

        project_folder = FmuBuilder._get_project_folder(temp_dest)
        deck_file = project_folder / DECK_FILE_NAME
        system.save(deck_file)  # Save the customized system

        # Create default environment if it is not provided
        if python_env is None:
            python_env = project_folder / "requirements.txt"

        python_env = Path(python_env)

        if not python_env.exists():
            logger.info("Create default package list.")

            package_list = {"cosapp": cosapp_version}

            python_env.write_text(
                "\n".join(
                    [f"{name}~={version}" for name, version in package_list.items()]
                )
            )

        doc_folder = FmuBuilder._get_documentation_folder(temp_dest)
        license_file = None
        if license is not None and license.endswith("Proprietary"):
            company = license[: -len("Proprietary")].strip()
            license_file = doc_folder / "licenses" / "license.txt"
            license_file.parent.mkdir(parents=True, exist_ok=True)
            license_file.write_text(
                f"Copyright © {THIS_YEAR} {company} - All Rights Reserved\n"
                f"Unauthorized copying and/or distribution of this source-code, via any medium "
                f"is strictly prohibited\nwithout the express permission of {company}.\n"
                "Proprietary and confidential"
            )

        return temp_dest

    @staticmethod
    def to_fmu(
        system: System,
        inputs: Iterable[str] = None,
        parameters: Iterable[str] = None,
        outputs: Iterable[str] = None,
        locals: Iterable[str] = None,
        time_integrator: Union[TimeIntegrator, str] = TimeIntegrator.RK4,
        non_linear_solver: Optional[AbstractSolver] = None,
        dest: Union[Path, str] = os.curdir,
        python_env: Union[Path, str, None] = None,
        project_files: Iterable[Union[Path, str]] = set(),
        version: Optional[str] = None,
        author: Optional[str] = None,
        description: Optional[str] = None,
        copyright: Optional[str] = "",
        license: Optional[str] = "",
        fmu_name_suffix: str = "",
    ) -> Path:
        """Export a system as CoSimulation FMU respecting FMI 2.0.

        Inputs, parameters, outputs and locals are by default taken from inputs,
        inwards, outputs and outwards variables of the CoSApp system, respectively. If you wish
        to set an empty list, set the corresponding argument with a empty dictionary.

        You can add a mathematical problem to the system by passing the proper
        non_linear_solver. If it contains a RunSingleCase, its equations will be
        passed to the FMU. 
        
        Parameters
        ----------
        system : System
            System to be exported
        inputs : Iterable[str], optional
            List of input variables with initial values; default None
        parameters : Iterable[str], optional
            List of parameter variables with initial values; default None
        outputs : Iterable[str], optional
            List of output variables with initial values; default None
        locals : Iterable[str], optional
            List of local variables with initial values; default None
        time_integrator : TimeIntegrator, optional
            Time integrator algorithm; default Runge-Kutta 4th order
        non_linear_solver : AbstractSolver, optional
            Non linear Driver to use for solving the system at a given instant
        dest : str or Path, optional
            Destination folder; default current directory
        python_env : str or Path, optional
            File listing the python dependency; default None
        project_files : Iterable of str or Path, optional
            List of additional files to be included in the FMU; default no additional files
        version : str, optional
            FMU version; default None
        author : str, optional
            FMU author; default None
        description : str, optional
            System description; default None
        license : str, optional
            FMU license; default ""
        copyright : str, optional
            FMU copyright; default ""
        fmu_name_suffix : str, optional
            FMU name suffix; default ""

        Returns
        -------
        pathlib.Path
            FMU file path object
        """
        dest = Path(dest)
        temp_dest = dest / "tmp"

        temp_dest = FmuBuilder.generate_fmu_facade(
            system,
            inputs,
            parameters,
            outputs,
            locals,
            time_integrator,
            non_linear_solver,
            temp_dest,
            python_env,
            version,
            author,
            description,
            copyright,
            license,
            fmu_name_suffix,
        )

        if PyFmuBuilder is None:
            logger.warning(
                "pythonfmu needs to be installed to package the simulation as FMU."
            )
            return dest

        project_folder = FmuBuilder._get_project_folder(temp_dest)
        project_files = set([Path(f) for f in project_files])
        project_files.update(project_folder.glob("*"))

        documentation_folder = FmuBuilder._get_documentation_folder(temp_dest)
        PyFmuBuilder.build_FMU(
            FmuBuilder._get_script_file(temp_dest, system, fmu_name_suffix),
            dest=dest,
            project_files=project_files,
            documentation_folder=documentation_folder,
            needsExecutionTool="false",
            canGetAndSetFMUstate="false",
        )

        shutil.rmtree(temp_dest)

        return dest / f"{type(system).__name__}.fmu"


to_fmu = FmuBuilder.to_fmu
